/**
 * The $P Point-Cloud Recognizer (JavaScript version)
 *
 * 	Radu-Daniel Vatavu, Ph.D.
 *	University Stefan cel Mare of Suceava
 *	Suceava 720229, Romania
 *	vatavu@eed.usv.ro
 *
 *	Lisa Anthony, Ph.D.
 *      UMBC
 *      Information Systems Department
 *      1000 Hilltop Circle
 *      Baltimore, MD 21250
 *      lanthony@umbc.edu
 *
 *	Jacob O. Wobbrock, Ph.D.
 * 	The Information School
 *	University of Washington
 *	Seattle, WA 98195-2840
 *	wobbrock@uw.edu
 *
 * The academic publication for the $P recognizer, and what should be 
 * used to cite it, is:
 *
 *	Vatavu, R.-D., Anthony, L. and Wobbrock, J.O. (2012).  
 *	  Gestures as point clouds: A $P recognizer for user interface 
 *	  prototypes. Proceedings of the ACM Int'l Conference on  
 *	  Multimodal Interfaces (ICMI '12). Santa Monica, California  
 *	  (October 22-26, 2012). New York: ACM Press, pp. 273-280.
 *
 * This software is distributed under the "New BSD License" agreement:
 *
 * Copyright (c) 2012, Radu-Daniel Vatavu, Lisa Anthony, and 
 * Jacob O. Wobbrock. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *    * Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    * Neither the names of the University Stefan cel Mare of Suceava, 
 *	University of Washington, nor UMBC, nor the names of its contributors 
 *	may be used to endorse or promote products derived from this software 
 *	without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Radu-Daniel Vatavu OR Lisa Anthony
 * OR Jacob O. Wobbrock BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
**/
package com.example.dollarstroke;

import java.util.ArrayList;
import java.util.Arrays;
import java.lang.Double;
import java.lang.Integer;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.FileOutputStream;

import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import org.xmlpull.v1.XmlPullParser;
import org.xmlpull.v1.XmlPullParserException;
import org.xmlpull.v1.XmlPullParserFactory;

//- DollarRecognizer class constants
public class DollarRecognizer 
{
	public  final static int	RECON_FLAG_USE_PROTRACTOR	= 0x01;
	public  final static int	RECON_FLAG_USE_ENDPOINTDIST	= 0x02;
	public  final static int	RECON_FLAG_USE_ANGLEDIR		= 0x04;

	private final static Point	mPointOrig	= new Point(0.0,0.0,0);

	private final static int	mNumPoints	= 64;
	private final static double	SquareSize	= 250.0;
	private final static double	Diagonal	= Math.sqrt((SquareSize * SquareSize) + (SquareSize * SquareSize));
	private final static double	HalfDiagonal	= 0.5 * Diagonal;
	private final static double	AngleRange	= Deg2Rad(45.0);
	private final static double	AnglePrecision	= Deg2Rad(2.0);
	private final static double	Phi		= 0.5 * (-1.0 + Math.sqrt(5.0)); // Golden Ratio

	private void DollarRecognizer(){}

	private static double CloudDistance(ArrayList<Point> pts1, ArrayList<Point> pts2, double start)
	{
		//- pts1.size() == pts2.size()
		double pointCount = (double)pts1.size();

		boolean[] matched = new boolean[pts1.size()]; 
		for( int k = 0; k < pts1.size(); k++ )
			matched[k] = false;
		double sum = 0.0;
		double i = start;
		do
		{
			int index = -1;
			double min = Double.POSITIVE_INFINITY;
			for( int j = 0; j < matched.length; j++ )
			{
				if( !matched[j] ) {
					double d = euclideanDistance(pts1.get((int)i), pts2.get(j));
					if( d < min ) {
						min = d;
						index = j;
					}
				}
			}
			matched[index] = true;
			double weight = 1.0 - ((i - start + pointCount) % pointCount) / pointCount;
			sum += weight * min;
			i = (i + 1.0) % pointCount;
		} while( i != start );
		return sum;
	}

	private static ArrayList<Point> Resample(ArrayList<Point> points, int n)
	{
		//- interval length
		double I = pathLength(points) / (double)(n - 1);
		double D = 0.0;

		ArrayList<Point> newpoints = new ArrayList<Point>(); 
		Point p = points.get(0);
		newpoints.add(new Point(p.X,p.Y,p.ID));

		for( int i = 1; i < points.size(); i++ )
		{
			Point pt1 = points.get(i);
			Point pt2 = points.get(i - 1);

			if( points.get(i).ID == points.get(i-1).ID )
			{
				double d = euclideanDistance(points.get(i - 1), points.get(i));
				if ((D + d) >= I)
				{
					double qx = pt2.X + ((I - D) / d) * (pt1.X - pt2.X);
					double qy = pt2.Y + ((I - D) / d) * (pt1.Y - pt2.Y);
					Point q = new Point(qx, qy, pt1.ID);
					newpoints.add(q); //- append new point 'q'
					points.add(i, q); //- insert 'q' at position i in points s.t. 'q' will be the next i
					D = 0.0;
				} else {
					D += d;
				}
			}
		}

		//- sometimes we fall a rounding-error short of
		//- adding the last point, so add it if so
		if( newpoints.size() == n - 1 ) 
		{ 
			p = points.get(points.size() - 1);
			newpoints.add(new Point(p.X,p.Y,p.ID));
		}
		return newpoints;
	}

	private static ArrayList<Point> scaleTo(ArrayList<Point> points, double size)
	{
		BoundingBox bBox = new BoundingBox(points);

		ArrayList<Point> newpoints = new ArrayList<Point>();
		for( Point point : points ) {
			double qx = point.X * (size / bBox.width);
			double qy = point.Y * (size / bBox.height);
			newpoints.add(new Point(qx, qy, point.ID));
		}
		return newpoints;
	}

	private static ArrayList<Point> scale(ArrayList<Point> points)
	{
		BoundingBox bBox = new BoundingBox(points);

		double size = Math.max(bBox.width,bBox.height);

		ArrayList<Point> newpoints = new ArrayList<Point>();
		for( Point point : points ) {
			double qx = (point.X - bBox.X) / size;
			double qy = (point.Y - bBox.Y) / size;
			newpoints.add(new Point(qx, qy, point.ID));
		}
		return newpoints;
	}

	private static double IndicativeAngle(ArrayList<Point> points)
	{
		Point c = Centroid(points);
		return Math.atan2(c.Y-points.get(0).Y, c.X-points.get(0).X);
	}

	private static double pointToPointAngle(Point pt1, Point pt2)
	{
		return Math.atan2(pt2.Y - pt1.Y, pt2.X -pt1.X) * 180.0 / Math.PI;
	}

	private static ArrayList<Point> rotateBy(ArrayList<Point> points, double radians)
	{
		Point c = Centroid(points);
		ArrayList<Point> newPoints = new ArrayList<Point>();
		double cos = Math.cos(radians);
		double sin = Math.sin(radians);
		for( Point point : points ) {
			double qx = (point.X - c.X)*cos - (point.Y - c.Y)*sin + c.X;
			double qy = (point.X - c.X)*sin + (point.Y - c.Y)*cos + c.Y;
			newPoints.add(new Point(qx, qy, point.ID));
		}
		return newPoints;
	}

	//- Translates points' centroid
	private static ArrayList<Point> TranslateTo(ArrayList<Point> points, Point pt)
	{
		Point c = Centroid(points);
		ArrayList<Point> newpoints = new ArrayList<Point>();
		for( Point point : points ) {
			double qx = point.X + pt.X - c.X;
			double qy = point.Y + pt.Y - c.Y;
			newpoints.add(new Point(qx, qy, point.ID));
		}
		return newpoints;
	}

	private static double[] Vectorize(ArrayList<Point> points)
	{
		double sum = 0.0;
		double[] vector = new double[points.size() * 2];
		int index = 0;
		for( Point point : points )
		{
			vector[index++] = point.X;
			vector[index++] = point.Y;
			sum += (point.X * point.X) + (point.Y * point.Y);
		}

		double magnitude = Math.sqrt(sum);
		for( int i = 0; i < vector.length; i++ )
			vector[i] /= magnitude;

		return vector;
	}

	public static double OptimalCosineDistance(double[] v1, double[] v2)
	{
		double a = 0.0;
		double b = 0.0;

		int len = Math.min(v1.length, v2.length);
		for( int i = 0; i < len; i += 2 )
		{
			a += v1[i] * v2[i] + v1[i + 1] * v2[i + 1];
			b += v1[i] * v2[i + 1] - v1[i + 1] * v2[i];
		}
		double angle = Math.atan(b / a);
		return Math.acos(a * Math.cos(angle) + b * Math.sin(angle));
	}

	public static double distanceAtBestAngle(
			ArrayList<Point> points,
			PointTemplate t,
			double aTheta,
			double bTheta,
			double dTheta
		)
	{
		double x1 = Phi*aTheta + (1.0-Phi)*bTheta;
		double f1 = distanceAtAngle(points, t, x1);
		double x2 = (1.0-Phi)*aTheta + Phi*bTheta;
		double f2 = distanceAtAngle(points, t, x2);

		while( Math.abs(bTheta-aTheta) > dTheta )
		{
			if(f1 < f2){
				bTheta = x2;
				x2 = x1;
				f2 = f1;
				x1 = Phi*aTheta + (1.0-Phi)*bTheta;
				f1 = distanceAtAngle(points, t, x1);
			} else {
				bTheta = x1;
				x1 = x2;
				f1 = f2;
				x2 = (1.0-Phi)*aTheta + Phi*bTheta;
				f2 = distanceAtAngle(points, t, x2);
			}
		}
		return Math.min(f1, f2);
	}

	private static double distanceAtAngle(ArrayList<Point> points, PointTemplate t, double theta)
	{
		ArrayList<Point> newPoints = rotateBy(points, theta);
		return pathDistance(newPoints, t.mPoints);
	}

	private static Point Centroid(ArrayList<Point> points)
	{
		double x = 0.0;
		double y = 0.0;
		double len = (double)points.size();
		for( Point point : points ) {
			x += point.X;
			y += point.Y;
		}
		return new Point(x/len, y/len, 0);
	}

	//- average distance between corresponding points in two paths
	private static double pathDistance(ArrayList<Point> pts1, ArrayList<Point> pts2)
	{
		double d = 0.0;
		//- assumes pts1.size() == pts2.size()
		for( int i = 0; i < pts1.size(); i++ ) 
			d += euclideanDistance(pts1.get(i), pts2.get(i));
		return d / (double)pts1.size();
	}

	//- Euclidean distance between two points
	private static double euclideanDistance(Point pt1, Point pt2)
	{
		double dx = pt2.X - pt1.X;
		double dy = pt2.Y - pt1.Y;
		return Math.sqrt(dx * dx + dy * dy);
	}

	//- length traversed by a point path
	private static double pathLength(ArrayList<Point> points)
	{
		double d = 0.0;
		for( int i = 1; i < points.size(); i++ )
		{
			Point pt1 = points.get(i);
			Point pt2 = points.get(i - 1);

			if( pt1.ID == pt2.ID )
				d += euclideanDistance(pt2,pt1);
		}
		return d;
	}

	public static double Deg2Rad(double d) { return (d * Math.PI / 180.0); }

	public static void writePointsToXml(FileOutputStream outputStream, PointTemplate pntTmpl)
		throws IOException
	{
		int strokeIndex = 1;
		outputStream.write(String.format(
			"<Gesture Name=\"%s\" asci=\"%d\" NumPts=\"%d\">\n",
			pntTmpl.mName, pntTmpl.mNumValue, pntTmpl.mPoints.size()).getBytes());

		outputStream.write(String.format("\t<Stroke index=\"%d\">\n",strokeIndex).getBytes());
		for( Point point : pntTmpl.mPoints )
		{
			if( point.ID == strokeIndex )
			{
				outputStream.write(
					String.format("\t\t<Point X=\"%3.0f\" Y=\"%3.0f\"/>\n",point.X,point.Y).getBytes());
			} else {
				outputStream.write(String.format("\t</Stroke>\n").getBytes());
				outputStream.write(String.format("\t<Stroke index=\"%d\">\n",point.ID).getBytes());
				outputStream.write(String.format(
					"\t\t<Point X=\"%3.0f\" Y=\"%3.0f\"/>\n", point.X,point.Y).getBytes());
				strokeIndex = point.ID;
			}
		}
		outputStream.write(String.format("\t</Stroke>\n").getBytes());
		outputStream.write(String.format("</Gesture>\n").getBytes());
	}

	static abstract class Recognizer
	{
		protected ArrayList<PointTemplate> mPntTemplate = new ArrayList<PointTemplate>();

		abstract void addPointTemplate(PointTemplate pntTemplate);

		abstract void addPointTemplate(String name, ArrayList<Point> points, int asci);

		abstract Results Recognize(ArrayList<Point> points, int configFlag);
		
		protected void initializefromXml()
		{
			try {
				SAXParserFactory factory = SAXParserFactory.newInstance();
				SAXParser saxParser = factory.newSAXParser();

				DefaultHandler handler = new DefaultHandler() {
					ArrayList<Point> points;
					String	name;
					int	stroke		= 0;
					int	asciValue	= 0;

					public void startElement(String uri, String localName, String qName, Attributes attributes)
						throws SAXException
					{
						if( qName.equalsIgnoreCase("Gesture") )
						{
							points = new ArrayList<Point>();
							name = attributes.getValue("Name");
							asciValue = Integer.parseInt(attributes.getValue("Asci"));
						}

						if( qName.equalsIgnoreCase("Stroke") )
						{
							stroke =  Integer.parseInt(attributes.getValue("index"));
						}

						if( qName.equalsIgnoreCase("Point") )
						{
							points.add(
								new Point(
									Double.parseDouble(attributes.getValue("X")),
									Double.parseDouble(attributes.getValue("Y")), stroke)
								);
						}
					}

					public void endElement(String uri, String localName, String qName)
						throws SAXException
					{
						if( qName.equalsIgnoreCase("Gesture") )
						{
							addPointTemplate(name, points, asciValue);
						}
					}
				};

				saxParser.parse("bank.xml", handler);

			} catch( Exception e ) {
				e.printStackTrace();
			}
		}

		protected void initializefromXmlAndorid(InputStream inputStream)
			throws XmlPullParserException
		{
			try {
				XmlPullParserFactory factory=XmlPullParserFactory.newInstance();
				factory.setNamespaceAware(true);
				XmlPullParser parser = factory.newPullParser();
		
				parser.setInput(inputStream,null);
				
				ArrayList<Point> points = null;
				String	name		= null;
				String	tagName		= null;
				int	stroke 		= 0;
				int	asciValue	= 0;

				for( int eventType = parser.getEventType();
						eventType != XmlPullParser.END_DOCUMENT;
									eventType = parser.next() )
				{
					switch( eventType )
					{
						case XmlPullParser.START_DOCUMENT:
							break;

						case XmlPullParser.START_TAG:
							tagName = parser.getName();
							if( tagName.equalsIgnoreCase("Gesture") )
							{
								points = new ArrayList<Point>();
								name = parser.getAttributeValue(null,"Name");
								asciValue = Integer.parseInt(parser.getAttributeValue(null,"asci"));
							}

							if( tagName.equalsIgnoreCase("Stroke") )
							{
								stroke =  Integer.parseInt(parser.getAttributeValue(null,"index"));
							}

							if( tagName.equalsIgnoreCase("Point") )
							{
								if( points != null )
									points.add(
										new Point(
											Double.parseDouble(parser.getAttributeValue(null,"X")),
											Double.parseDouble(parser.getAttributeValue(null,"Y")),
											stroke)
										);
							}
							break;

						case XmlPullParser.END_TAG:
							tagName = parser.getName();
							if( tagName.equalsIgnoreCase("Gesture") )
							{
								addPointTemplate(name, points, asciValue);
							}
							break;
					}
				}
			} catch( Exception e ) {
				e.printStackTrace();
			}
		}
	}

	public static class UniRecognizer extends Recognizer
	{
		public UniRecognizer(InputStream inputStream)
			throws XmlPullParserException
		{
			initializefromXmlAndorid(inputStream);
		}

		public Results Recognize(ArrayList<Point> points, int configFlag)
		{
			points = Resample(points, mNumPoints);
			double radians = IndicativeAngle(points);
			points = rotateBy(points, -radians);
			points = scaleTo(points, SquareSize);
			points = TranslateTo(points, mPointOrig);
			double[] vector = Vectorize(points); // for Protractor

			double score = Double.POSITIVE_INFINITY;
			PointTemplate foundPointTemplate = null;
			for( PointTemplate pntTmpl : mPntTemplate ) 
			{
				double distScore;
				if( (configFlag & RECON_FLAG_USE_PROTRACTOR) == RECON_FLAG_USE_PROTRACTOR ) 
					{ // for Protractor
					distScore = OptimalCosineDistance(pntTmpl.mVector, vector);
				} else { // Golden Section Search (original $1)
					distScore = distanceAtBestAngle(points, pntTmpl, -AngleRange, +AngleRange, AnglePrecision);
				}
				if( distScore < score ) {
					score = distScore; // best (least) distance
					foundPointTemplate = pntTmpl; // unistroke
				}
			}

			double wScore = (configFlag & RECON_FLAG_USE_PROTRACTOR) == RECON_FLAG_USE_PROTRACTOR  
					? 1.0 / score : 1.0 - score / HalfDiagonal;
			return new Results(foundPointTemplate, wScore, 0.0);
		}

		public void addPointTemplate(String name, ArrayList<Point> points, int asci)
		{
			PointTemplate pntT = new PointTemplate(name, points, asci);
			pntT.uniDNormalizeTemplate();
			mPntTemplate.add(pntT);
		}

		public void addPointTemplate(PointTemplate pntT)
		{
			pntT.pDNormalizeTemplate();
			mPntTemplate.add(pntT);
		}
	}

        public static class PRecognizer extends Recognizer
	{
		public PRecognizer(InputStream inputStream)
			throws XmlPullParserException
		{
			initializefromXmlAndorid(inputStream);
		}

		public Results Recognize(ArrayList<Point> points, int configFlag)
		{
			// the following is NOT part of the originally published javascript implementation
			// and has been added to support addition of directional testing for point clouds
			// which represent unistroke gestures

			PointTemplate testTmpl = new PointTemplate(points);
			testTmpl.pDNormalizeTemplate();

			double angle = 0.0;
			double foundAngle = 0.0;
			double score = Double.POSITIVE_INFINITY;
			PointTemplate foundPointTemplate = null;
			for( PointTemplate pntTmpl : mPntTemplate ) 
			{
				if( (configFlag & RECON_FLAG_USE_ANGLEDIR) == RECON_FLAG_USE_ANGLEDIR )
				{
					// test to see if the gestures match roughly directionality
					// if not, keep looking
					angle = pntTmpl.vectorDirectionAngle(testTmpl);
					if( angle > 25.0 )
					{
						continue;
					}
				}

				// directional testing for point clouds which represent unistroke gestures
				if( (configFlag & RECON_FLAG_USE_ENDPOINTDIST) == RECON_FLAG_USE_ENDPOINTDIST )
				{
					// test to see if the gestures match roughly
					// in directionality if not, keep looking
					if( testTmpl.distantBetweenEndPoint(pntTmpl,.3,.3) )
					{
						continue;
					}
				}

				double distScore = testTmpl.GreedyCloudMatch(pntTmpl);
				if( distScore < score ) 
				{
					score = distScore; //- best (least) distance
					foundPointTemplate = pntTmpl; //- point-Template
					foundAngle = angle;
				}
			}

			double wScore = Math.max((score - 2.0) / -2.0, 0.0);
			return new Results(foundPointTemplate, wScore, foundAngle);
		}

		public void addPointTemplate(String name, ArrayList<Point> points, int asci)
		{
			PointTemplate pntT = new PointTemplate(name, points, asci);
			pntT.pDNormalizeTemplate();
			mPntTemplate.add(pntT);
		}

		public void addPointTemplate(PointTemplate pntT)
		{
			pntT.pDNormalizeTemplate();
			mPntTemplate.add(pntT);
		}
	}

	//- PointTemplate class: a point-Template template
	public static class PointTemplate
	{
		public ArrayList<Point> mPoints;
		protected LineVector 	mDirLineVector;
		public double[]		mVector;
		public String		mName;
		public int		mNumValue = 0;
		protected boolean	mNormalized = false;
		protected boolean	mIsUnistroke = true;
		protected double	mDirectionAngle = 0.0;

		public PointTemplate(ArrayList<Point> points)
		{
			mPoints		= points;

			int id = points.get(0).ID;
			for( Point point : points )
			{
				if( point.ID != id )
				{
					mIsUnistroke = false;
					break;
				}
			}
			mDirLineVector = new LineVector(mPoints);
		}

		public PointTemplate(String name, ArrayList<Point> points, int asci)
		{
			mName		= name;
			mPoints		= points;
			mNumValue	= asci;
			
			int id = points.get(0).ID;
			for( Point point : points )
			{
				if( point.ID != id )
				{
					mIsUnistroke = false;
					break;
				}
			}
			mDirLineVector = new LineVector(mPoints);
		}

		protected double vectorDirectionAngle(PointTemplate tmpl)
		{
			return mDirLineVector.angleBetweenVector(tmpl.mDirLineVector);
		}

		protected double GreedyCloudMatch(PointTemplate pntTmpl)
		{
			double e = 0.50;
			double step = Math.floor(Math.pow((double)mPoints.size(), 1 - e));

			double min = Double.POSITIVE_INFINITY;
			for( double i = 0.0; i < mPoints.size(); i += step )
			{
				double d1 = CloudDistance(mPoints, pntTmpl.mPoints, i);
				double d2 = CloudDistance(pntTmpl.mPoints, mPoints, i);
				min = Math.min(min, Math.min(d1, d2));
			}
			return min;
		}

		protected boolean distantBetweenEndPoint(PointTemplate tmpl, double startRange, double endRange)
		{
			Point refStart = tmpl.mPoints.get(0);
			Point inStart  = mPoints.get(0);
			Point refEnd   = tmpl.mPoints.get(tmpl.mPoints.size()-1);
			Point inEnd    = mPoints.get(mPoints.size()-1);

			return ((euclideanDistance(refStart, inStart) > startRange) || (euclideanDistance(refEnd, inEnd) > endRange)); 
		}

		protected void pDNormalizeTemplate()
		{
			if( mNormalized == false )
			{
				mPoints = Resample(mPoints, mNumPoints);
				mPoints = scale(mPoints);
				mPoints = TranslateTo(mPoints, mPointOrig);
				mNormalized = true;
			}
		}

		protected void uniDNormalizeTemplate()
		{
			if( mNormalized == false )
			{
				mPoints = Resample(mPoints, mNumPoints);
				double radians = IndicativeAngle(mPoints);
				mPoints = rotateBy(mPoints, -radians);
				mPoints = scale(mPoints);
				mPoints = TranslateTo(mPoints, mPointOrig);
				mVector = Vectorize(mPoints); // for Protractor
				mNormalized = true;
			}
		}
	}

	private static class BoundingBox
	{
		public double X = Double.POSITIVE_INFINITY;
		public double Y = Double.POSITIVE_INFINITY;
		public double width;
		public double height;

		public BoundingBox(ArrayList<Point> points)
		{
			double maxX = Double.NEGATIVE_INFINITY;
			double maxY = Double.NEGATIVE_INFINITY;

			for( Point point : points ) {
				X = Math.min(X, point.X);
				Y = Math.min(Y, point.Y);
				maxX = Math.max(maxX, point.X);
				maxY = Math.max(maxY, point.Y);
			}

			width  = maxX - X;
			height = maxY - Y;
		}
	}

	//- Point class: 
	public static class Point
	{
		public double X;
		public double Y;
		public int ID;

		public Point(double x, double y, int id)
			{ X = x; Y = y; ID = id; }
	}

	private static class LineVector
	{
		public Point  vectDir;
		public double magnitude;
		public double vectorAngle;

		public LineVector(ArrayList<Point> point)
		{
			Point pt1	= point.get(0);
			Point pt2	= point.get(point.size()-1);
			vectDir		= new Point(pt1.X-pt2.X, pt1.Y-pt2.Y, 0);
			vectorAngle	= pointToPointAngle(pt1, pt2);
			magnitude	= Math.sqrt(Math.pow(vectDir.X,2.0)+Math.pow(vectDir.Y,2.0));
		}

		public double angleBetweenVector(LineVector vec)
		{
			double dotX = vec.vectDir.X * vectDir.X;
			double dotY = vec.vectDir.Y * vectDir.Y;

			return Math.acos((dotX+dotY) / (vec.magnitude * magnitude)) * 180.0 / Math.PI;
		}
	}

	//- Results class: 
	public static class Results
	{
		public PointTemplate	mPntTemplate;
		public ArrayList<Point> mPoints;
		public double		mScore = 0.0;
		public double		angle  = 0.0;

		public Results(PointTemplate pc, double s, double a) { mPntTemplate = pc; mScore = s; angle = a; }
	}
}
